# ------------------ Rust Downsamplers ------------------
from tsdownsample._rust import _tsdownsample_rs  # type: ignore[attr-defined]

from .downsampling_interface import RustDownsamplingInterface

MinMaxDownsampler = RustDownsamplingInterface("MinMax", _tsdownsample_rs.minmax)
M4Downsampler = RustDownsamplingInterface("M4", _tsdownsample_rs.m4)
LTTBDownsampler = RustDownsamplingInterface("LTTB", _tsdownsample_rs.lttb)
MinMaxLTTBDownsampler = RustDownsamplingInterface(
    "MinMaxLTTB", _tsdownsample_rs.minmaxlttb
)

# ------------------ Function Downsamplers ------------------
import numpy as np

from .downsampling_interface import FuncDownsamplingInterface

MeanDownsampler = FuncDownsamplingInterface("Mean", np.mean)
MedianDownsampler = FuncDownsamplingInterface("Median", np.median)

# ------------------ EveryNth Downsampler ------------------
import math

import pandas as pd

from .downsampling_interface import DownsampleInterface


class _EveryNthDownsampler(DownsampleInterface):
    def __init__(self) -> None:
        super().__init__("EveryNth")

    def downsample(self, s: pd.Series, n_out: int, parallel: bool = False) -> pd.Series:
        return s[:: max(1, math.ceil(len(s) / n_out))]


EveryNthDownsampler = _EveryNthDownsampler()



import warnings
from abc import ABC, abstractmethod
from types import ModuleType
from typing import Callable

import numpy as np
import pandas as pd


class DownsampleInterface(ABC):
    def __init__(self, name: str) -> None:
        self.name = name

    @staticmethod
    def _construct_output_series(s: pd.Series, idxs: np.ndarray) -> pd.Series:
        return s.iloc[idxs]

    # def _supports_dtype(self, s: pd.Series):
    #     # base case
    #     if self.dtype_regex_list is None:
    #         return

    #     for dtype_regex_str in self.dtype_regex_list:
    #         m = re.compile(dtype_regex_str).match(str(s.dtype))
    #         if m is not None:  # a match is found
    #             return
    #     raise ValueError(
    #         f"{s.dtype} doesn't match with any regex in {self.dtype_regex_list}"
    #     )

    @abstractmethod
    def downsample(self, s: pd.Series, n_out: int, parallel: bool = False) -> pd.Series:
        """Downsample a pandas series to n_out samples.

        Parameters
        ----------
        s : pd.Series
            The series to downsample.
        n_out : int
            The number of samples to downsample to.
        parallel : bool, optional
            Whether to use parallel processing (if available), by default False.

        Returns
        -------
        pd.Series
            The downsampled series.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.name}"


# ------------------- Rust Downsample Interface -------------------
DOWNSAMPLE_F = "downsample"


def _switch_mod_with_y(
    y_dtype: np.dtype, mod: ModuleType, downsample_func: str = DOWNSAMPLE_F
) -> Callable:
    """The x-data is not considered in the downsampling

    Assumes equal binning.

    Parameters
    ----------
    y_dtype : np.dtype
        The dtype of the y-data
    mod : ModuleType
        The module to select the appropriate function from
    downsample_func : str, optional
        The name of the function to use, by default DOWNSAMPLE_FUNC.
    """
    # FLOATS
    if np.issubdtype(y_dtype, np.floating):
        if y_dtype == np.float16:
            return getattr(mod, downsample_func + "_f16")
        elif y_dtype == np.float32:
            return getattr(mod, downsample_func + "_f32")
        elif y_dtype == np.float64:
            return getattr(mod, downsample_func + "_f64")
    # UINTS
    elif np.issubdtype(y_dtype, np.unsignedinteger):
        if y_dtype == np.uint16:
            return getattr(mod, downsample_func + "_u16")
        elif y_dtype == np.uint32:
            return getattr(mod, downsample_func + "_u32")
        elif y_dtype == np.uint64:
            return getattr(mod, downsample_func + "_u64")
    # INTS (need to be last because uint is subdtype of int)
    elif np.issubdtype(y_dtype, np.integer):
        if y_dtype == np.int16:
            return getattr(mod, downsample_func + "_i16")
        elif y_dtype == np.int32:
            return getattr(mod, downsample_func + "_i32")
        elif y_dtype == np.int64:
            return getattr(mod, downsample_func + "_i64")
    # BOOLS
    # TODO: support bools
    # elif data_dtype == np.bool:
    # return mod.downsample_bool
    raise ValueError(f"Unsupported data type (for y): {y_dtype}")


def _switch_mod_with_x_and_y(
    x_dtype: np.dtype, y_dtype: np.dtype, mod: ModuleType
) -> Callable:
    """The x-data is considered in the downsampling

    Assumes equal binning.

    Parameters
    ----------
    x_dtype : np.dtype
        The dtype of the x-data
    y_dtype : np.dtype
        The dtype of the y-data
    mod : ModuleType
        The module to select the appropriate function from
    """
    # FLOATS
    if np.issubdtype(x_dtype, np.floating):
        if x_dtype == np.float16:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_f16")
        elif x_dtype == np.float32:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_f32")
        elif x_dtype == np.float64:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_f64")
    # UINTS
    elif np.issubdtype(x_dtype, np.unsignedinteger):
        if x_dtype == np.uint16:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_u16")
        elif x_dtype == np.uint32:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_u32")
        elif x_dtype == np.uint64:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_u64")
    # INTS (need to be last because uint is subdtype of int)
    elif np.issubdtype(x_dtype, np.integer):
        if x_dtype == np.int16:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_i16")
        elif x_dtype == np.int32:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_i32")
        elif x_dtype == np.int64:
            return _switch_mod_with_y(y_dtype, mod, f"{DOWNSAMPLE_F}_i64")
    # BOOLS
    # TODO: support bools
    # elif data_dtype == np.bool:
    # return mod.downsample_bool
    raise ValueError(f"Unsupported data type (for x): {x_dtype}")


class RustDownsamplingInterface(DownsampleInterface):
    def __init__(self, name: str, resampling_mod: ModuleType) -> None:
        super().__init__(name + " [tsdownsample_rs]")
        self.rust_mod = resampling_mod

        # Store the single core sub module
        self.mod_single_core = self.rust_mod.scalar
        if hasattr(self.rust_mod, "simd"):
            # use SIMD implementation if available
            self.mod_single_core = self.rust_mod.simd

        # Store the multi-core sub module (if present)
        self.mod_multi_core = None  # no multi-core implementation (default)
        if hasattr(self.rust_mod, "simd_parallel"):
            # use SIMD implementation if available
            self.mod_multi_core = self.rust_mod.simd_parallel
        elif hasattr(self.rust_mod, "scalar_parallel"):
            # use scalar implementation if available (when no SIMD available)
            self.mod_multi_core = self.rust_mod.scalar_parallel

    def _downsample_without_x(self, s: pd.Series, n_out: int) -> pd.Series:
        downsample_method = _switch_mod_with_y(s.dtype, self.mod_single_core)
        idxs = downsample_method(s.values, n_out)
        return self._construct_output_series(s, idxs)

    def _downsample_with_x(self, s: pd.Series, n_out: int) -> pd.Series:
        downsample_method = _switch_mod_with_x_and_y(
            s.index.dtype, s.dtype, self.mod_single_core
        )
        idxs = downsample_method(s.index.values, s.values, n_out)
        return self._construct_output_series(s, idxs)

    def _downsample_without_x_parallel(self, s: pd.Series, n_out: int) -> pd.Series:
        if self.mod_multi_core is not None:
            downsample_method = _switch_mod_with_y(s.dtype, self.mod_multi_core)
        else:
            warnings.warn(
                "No multi-core implementation available. "
                "Falling back to single core implementation."
            )
            downsample_method = _switch_mod_with_y(s.dtype, self.mod_single_core)
        idxs = downsample_method(s.values, n_out)
        return self._construct_output_series(s, idxs)

    def _downsample_with_x_parallel(self, s: pd.Series, n_out: int) -> pd.Series:
        if self.mod_multi_core is not None:
            downsample_method = _switch_mod_with_x_and_y(
                s.index.dtype, s.dtype, self.mod_multi_core
            )
        else:
            warnings.warn(
                "No multi-core implementation available. "
                "Falling back to single core implementation."
            )
            downsample_method = _switch_mod_with_x_and_y(
                s.index.dtype, s.dtype, self.mod_single_core
            )
        idxs = downsample_method(s.index.values, s.values, n_out)
        return self._construct_output_series(s, idxs)

    def downsample(self, s: pd.Series, n_out: int, parallel: bool = False) -> pd.Series:
        fixed_sr = False
        if isinstance(s.index, pd.RangeIndex) or s.index.freq is not None:
            fixed_sr = True
        if fixed_sr:  # TODO: or the other way around??
            if parallel:
                return self._downsample_without_x_parallel(s, n_out)
            else:
                return self._downsample_without_x(s, n_out)
        else:
            if parallel:
                return self._downsample_with_x_parallel(s, n_out)
            else:
                return self._downsample_with_x(s, n_out)


# ------------------ Numpy Downsample Interface ------------------


class FuncDownsamplingInterface(DownsampleInterface):
    def __init__(self, name: str, downsample_func: Callable) -> None:
        super().__init__("[Func]_" + name)
        self.downsample_func = downsample_func

    def downsample(self, s: pd.Series, n_out: int, parallel: bool = False) -> pd.Series:
        if isinstance(s.index, pd.DatetimeIndex):
            t_start, t_end = s.index[:: len(s) - 1]
            rate = (t_end - t_start) / n_out
            return s.resample(rate).apply(self.downsample_func).dropna()

        # no time index -> use the every nth heuristic
        group_size = max(1, np.ceil(len(s) / n_out))
        s_out = (
            s.groupby(
                # create an array of [0, 0, 0, ...., n_out, n_out]
                # where each value is repeated based $len(s)/n_out$ times
                by=np.repeat(np.arange(n_out), group_size)[: len(s)]
            )
            .agg(self.downsample_func)
            .dropna()
        )
        # Create an index-estimation for real-time data
        # Add one to the index so it's pointed at the end of the window
        # Note: this can be adjusted to .5 to center the data
        # Multiply it with the group size to get the real index-position
        # TODO: add option to select start / middle / end as index
        idx_locs = (np.arange(len(s_out)) + 1) * group_size
        idx_locs[-1] = len(s) - 1
        return pd.Series(
            index=s.iloc[idx_locs.astype(s.index.dtype)].index.astype(s.index.dtype),
            data=s_out.values,
            name=str(s.name),
            copy=False,
        )



import tsdownsample._rust._tsdownsample_rs as tsds_rs

DTYPES = ["f16", "f32", "f64", "i16", "i32", "i64", "u16", "u32", "u64"]


def _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl: bool):
    # Without x
    for sub_mod in sub_mods:
        assert hasattr(mod, sub_mod)
        m = getattr(mod, sub_mod)
        for ty in DTYPES:
            assert hasattr(m, f"downsample_{ty}")
    # With x
    if not has_x_impl:
        return
    for sub_mod in sub_mods:
        assert hasattr(mod, sub_mod)
        m = getattr(mod, sub_mod)
        for tx in DTYPES:
            for ty in DTYPES:
                assert hasattr(m, f"downsample_{tx}_{ty}")


def test_minmax_rust_mod_correctly_build():
    mod = tsds_rs.minmax
    sub_mods = ["simd", "simd_parallel", "scalar", "scalar_parallel"]
    _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl=False)


def test_m4_rust_mod_correctly_build():
    mod = tsds_rs.m4
    sub_mods = ["simd", "simd_parallel", "scalar", "scalar_parallel"]
    _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl=False)


def test_lttb_rust_mod_correctly_build():
    mod = tsds_rs.lttb
    sub_mods = ["scalar"]
    _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl=True)


def test_minmaxlttb_rust_mod_correctly_build():
    mod = tsds_rs.minmaxlttb
    sub_mods = ["simd", "simd_parallel", "scalar", "scalar_parallel"]
    _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl=True)



import numpy as np
import pandas as pd

from tsdownsample import (
    EveryNthDownsampler,
    LTTBDownsampler,
    M4Downsampler,
    MeanDownsampler,
    MedianDownsampler,
    MinMaxDownsampler,
    MinMaxLTTBDownsampler,
)

# Very basic & poor tests
# TODO: Improve tests
#   - test with and without x
#   - test with and without parallel
#   - test with different data types
#   - test with different downsamplers
#   - compare implementations with existing plotly_resampler implementations


def test_m4_downsampler():
    """Test M4 downsampler."""
    arr = np.array(np.arange(10_000))
    s = pd.Series(arr)
    s_downsampled = M4Downsampler.downsample(s, 100)
    assert s_downsampled.values[0] == 0
    assert s_downsampled.values[-1] == len(arr) - 1


def test_minmax_downsampler():
    """Test MinMax downsampler."""
    arr = np.array(np.arange(10_000))
    s = pd.Series(arr)
    s_downsampled = MinMaxDownsampler.downsample(s, 100)
    assert s_downsampled.values[0] == 0
    assert s_downsampled.values[-1] == len(arr) - 1


def test_lttb_downsampler():
    """Test LTTB downsampler."""
    arr = np.array(np.arange(10_000))
    s = pd.Series(arr)
    s_downsampled = LTTBDownsampler.downsample(s, 100)
    assert s_downsampled.values[0] == 0
    assert s_downsampled.values[-1] == len(arr) - 1


def test_minmaxlttb_downsampler():
    """Test MinMaxLTTB downsampler."""
    arr = np.array(np.arange(10_000))
    s = pd.Series(arr)
    s_downsampled = MinMaxLTTBDownsampler.downsample(s, 100)
    assert s_downsampled.values[0] == 0
    assert s_downsampled.values[-1] == len(arr) - 1


def test_mean_downsampler():
    """Test Mean downsampler."""
    arr = np.array(np.arange(10_000))
    s = pd.Series(arr)
    s_downsampled = MeanDownsampler.downsample(s, 100)
    assert s_downsampled.values[0] == 49.5
    assert s_downsampled.values[-1] == 9_949.5


def test_median_downsampler():
    """Test Median downsampler."""
    arr = np.array(np.arange(10_000))
    s = pd.Series(arr)
    s_downsampled = MedianDownsampler.downsample(s, 100)
    assert s_downsampled.values[0] == 49.5
    assert s_downsampled.values[-1] == 9_949.5


def test_everynth_downsampler():
    """Test EveryNth downsampler."""
    arr = np.array(np.arange(10_000))
    s = pd.Series(arr)
    s_downsampled = EveryNthDownsampler.downsample(s, 100)
    assert s_downsampled.values[0] == 0
    assert s_downsampled.values[-1] == 9_900


## Parallel downsampling

rust_downsamplers = [
    MinMaxDownsampler,
    M4Downsampler,
    LTTBDownsampler,
    MinMaxLTTBDownsampler,
]


def test_parallel_downsampling():
    """Test parallel downsampling."""
    arr = np.random.randn(10_000).astype(np.float32)
    s = pd.Series(arr)
    for downsampler in rust_downsamplers:
        s_downsampled = downsampler.downsample(s, 100, parallel=False)
        s_downsampled_p = downsampler.downsample(s, 100, parallel=True)
        assert s_downsampled.equals(s_downsampled_p)


## Data types

supported_dtypes = [
    np.float16,
    np.float32,
    np.float64,
    np.int16,
    np.int32,
    np.int64,
    np.uint16,
    np.uint32,
    np.uint64,
]


def test_downsampling_different_dtypes():
    """Test downsampling with different data types."""
    arr = np.random.randn(10_000)
    for dtype in supported_dtypes:
        s = pd.Series(arr.astype(dtype))
        s_downsampled = MinMaxDownsampler.downsample(s, 100)
        assert s_downsampled.dtype == dtype
